/* Clase Organismo

Exones: son las porciones de ADN con un cdigo funcional.
Se ha argumentado que los exones podran ser antiguos "minigenes"
que en un principio codificaban pequeas secuencias funcionales
de aminocidos. Segn esta hiptesis, las protenas modernas
habran surgido por agregacin de varias de estas secuencias
aminocidos.

Intrones: En muchos organismos, una cantidad sorprendentemente grande
del ADN no parece intervenir activamente en la codificacin de
proteinas. Esta proporcin vara segn las especies: por ejemplo del 
9% al 27% en humanos, 75% en nematodos, 67% en moscas de la fruta,
99% en peces pulmonados. Parte de este ADN no codificador esta formado
por intrones.

Esta clase representara un exon o intron porque es en si una
instruccin completa del algoritmo gentico.

En el ADN de cualquier ser vivo se encuentra cuatro(4) tipos de bases
nitrogenadas que forman los nucletidos:
Adenina (A)
Timina (T)
Citosina (C)
Guanina (G)

Cada peldao de la escalera del ADN esta formado por la unin de dos
bases, la pareja es as:
Adenina con Timina
Citosina con Guanina

Los nucletidos A, T, C, G son como las letras de un alfabeto gentico,
con el que se ha elaborado un lenguaje para la informacin gentica.
Este lenguaje esta formado exclusivamente por tres palabras de tres
letras: AAA, CGT, TCC, etc.. Esto es lo que se conoce como cdigo de
tripletes. En ltimo trmino, cada uno de estos tripletes equivale,
en trminos de produccin, a un aminocido concreto. Los aminocidos
son las unidades estructurales de las protenas; las protenas forman
el cuerpo de los organismos, y tambin son protenas las enzimas y
hormonas responsables de su funcionamiento.

Autor:		Rafael Alberto Moreno Parra
URL:		http://www.geocities.com/krousky
Clase:		Organismo
Lenguaje:	C++
Fecha:		01 de Julio de 2002

*/

#include <time.h>
#include "Aleatorio.h"

class Organismo {
private:
	struct stTriplete
	{
		bool bEjecuta; /* Se coloca a TRUE si la instruccion es ejecutada */
		unsigned int iInstruccion; /* Cdigo de la Instruccin IF_Mayor, IF_Menor, SET o una funcion */
		unsigned int iVariable; /* Variable para asignarse o para comparacin */
		bool bEsVariable1; /* iOperando1 es una variable o un numero */
		unsigned int iOperando1; /* Variable o numero 1 en la expresin */
		bool bEsVariable2; /* iOperando2 es una variable o un numero */
		unsigned int iOperando2; /* Variable o numero 1 en la expresin */
		unsigned char cOperador; /* + - por dividido */
		unsigned int iGotoLabel; /* Hacia que instruccin ir */
	}; /* 1bit + 4bytes + 4bytes + 1bit + 4bytes + 1bit + 4bytes + 1byte + 4 bytes = 22 bytes */
	/* Nota: El anterior consumia 86 bytes por instruccin */

	bool m_bGenRandom; //Nmero de Genes fijos o aleatorio
	unsigned int m_iMinInstr; //Minimo numero de intrucciones
	unsigned int m_iMaxInstr; //Maximo numero de intrucciones
	unsigned int m_iTotalFuncion; //Total Funciones del ambiente
    unsigned int m_iTotalVariables; //Total Variables del Organismo
  	unsigned int m_iMaxiCiclos; //Maximos ciclos para evaluar organismo
	unsigned int m_iProbabIF; //Probabilidad IF
	unsigned int m_iProbabFN; //Probabilidad Funcion
	unsigned int m_iProbaVar; //Probabilidad de usar Variables

public:
    //Constructor
	Organismo();

	//Inicia la semilla de los nmeros aleatorios (tiempo)
	Aleatorio objAzar;
	void IniciaSemilla(unsigned long int iSemilla);
	
	//Fija las probabilidades en la estructura
	void vInicio (bool bGenRandom,
				  unsigned int iMinInstr,
				  unsigned int iMaxInstr,
                  unsigned int iTotalFuncion,
				  unsigned int iTotalVariable,
  				  unsigned int iMaxiCiclos,
				  unsigned int iProbabIF,
				  unsigned int iProbabFN,
				  unsigned int iProbaVar);

	//Crea el Organismo
    void vCreaADN();
    void vHaceGen(unsigned int iLabel);
	void vMutaGen();


	//Muestra como esta armado el Organismo
	void sDisplayADN(char *sbADN);

	//Ejecuta el Organismo
	int fEvalOrganismo (unsigned int *iOrgGenInst, unsigned int *iFuncion);
	float fVariable[70]; // Hasta 70 variables

	struct stTriplete m_oTriplete[5000]; //Hasta 5000 instrucciones
	unsigned int m_iMaxGenOrg; //Maximo numero de intrucciones real
	unsigned int iPuntAdapta; //Adaptacin ser vivo

	int _matherr(struct exception *a);
};
